@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2019, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -519,7 +519,8 @@ exports.modify_headers = function(task, hdr_alterations)
   local remove = hdr_alterations.remove or {}
 
   local hdr_flattened = {} -- For C API
-  for hname,hdr in pairs(add) do
+
+  local function flatten_add_header(hname, hdr)
     if not hdr_flattened[hname] then
       hdr_flattened[hname] = {add = {}}
     end
@@ -532,6 +533,17 @@ exports.modify_headers = function(task, hdr_alterations)
       end
     end
   end
+  if hdr_alterations.order then
+    -- Get headers alterations ordered
+    for _,hname in ipairs(hdr_alterations.order) do
+      flatten_add_header(hname, add[hname])
+    end
+  else
+    for hname,hdr in pairs(add) do
+      flatten_add_header(hname, hdr)
+    end
+  end
+
 
   for hname,hdr in pairs(remove) do
     if not hdr_flattened[hname] then
@@ -560,4 +572,145 @@ exports.modify_headers = function(task, hdr_alterations)
   end
 end
 
+--[[[
+-- @function lua_mime.message_to_ucl(task, [stringify_content])
+-- Exports a message to an ucl object
+--]]
+exports.message_to_ucl = function(task, stringify_content)
+  local E = {}
+
+  local maybe_stringify_f = stringify_content and
+    tostring or function(t) return t  end
+  local result = {
+    size = task:get_size(),
+    digest = task:get_digest(),
+    newlines = task:get_newlines_type(),
+    headers = task:get_headers(true)
+  }
+
+  -- Utility to convert ip addr to a string or nil if invalid/absent
+  local function maybe_stringify_ip(addr)
+    if addr and addr:is_valid() then
+      return addr:to_string()
+    end
+
+    return nil
+  end
+  -- Envelope (smtp) information form email
+  result.envelope = {
+    from_smtp = (task:get_from('smtp') or E)[1],
+    recipients_smtp = task:get_recipients('smtp'),
+    helo = task:get_helo(),
+    hostname = task:get_hostname(),
+    client_ip = maybe_stringify_ip(task:get_client_ip()),
+    from_ip = maybe_stringify_ip(task:get_from_ip()),
+  }
+
+  local parts = task:get_parts() or E
+  result.parts = {}
+  for _,part in ipairs(parts) do
+    if not part:is_multipart() and not part:is_message() then
+      local p = {
+        size = part:get_length(),
+        type = string.format('%s/%s', part:get_type()),
+        detected_type = string.format('%s/%s', part:get_detected_type()),
+        filename = part:get_filename(),
+        content = maybe_stringify_f(part:get_content()),
+        headers =  part:get_headers(true) or E,
+        boundary = part:get_enclosing_boundary(),
+      }
+      table.insert(result.parts, p)
+    else
+      -- Service part: multipart container or message/rfc822
+      local p = {
+        type = string.format('%s/%s', part:get_type()),
+        headers =  part:get_headers(true) or E,
+        boundary = part:get_enclosing_boundary(),
+        size = 0,
+      }
+
+      if part:is_multipart() then
+        p.multipart_boundary = part:get_boundary()
+      end
+
+      table.insert(result.parts, p)
+    end
+  end
+
+  return result
+end
+
+--[[[
+-- @function lua_mime.message_to_ucl_schema()
+-- Returns schema for a message to verify result/document fields
+--]]
+exports.message_to_ucl_schema = function()
+  local ts = require("tableshape").types
+
+  local function headers_schema()
+    return ts.shape{
+      order =  ts.integer:describe('Header order in a message'),
+      raw = ts.string:describe('Raw header value'):is_optional(),
+      empty_separator = ts.boolean:describe('Whether header has an empty separator'),
+      separator = ts.string:describe('Separator between a header and a value'),
+      decoded = ts.string:describe('Decoded value'):is_optional(),
+      value = ts.string:describe('Decoded value'):is_optional(),
+      name = ts.string:describe('Header name'),
+      tab_separated = ts.boolean:describe('Whether header has tab as a separator')
+    }
+  end
+
+  local function part_schema()
+    return ts.shape{
+      content =  ts.string:describe('Decoded content'):is_optional(),
+      multipart_boundary = ts.string:describe('Multipart service boundary'):is_optional(),
+      size = ts.integer:describe('Size of the part'),
+      type = ts.string:describe('Announced type'):is_optional(),
+      detected_type = ts.string:describe('Detected type'):is_optional(),
+      boundary = ts.string:describe('Eclosing boundary'):is_optional(),
+      filename = ts.string:describe('File name for attachments'):is_optional(),
+      headers = ts.array_of(headers_schema()):describe('Part headers'),
+    }
+  end
+
+  local function email_addr_schema()
+    return ts.shape{
+      addr =  ts.string:describe('Parsed address'):is_optional(),
+      raw = ts.string:describe('Raw address'),
+      flags = ts.shape{
+        valid = ts.boolean:describe('Valid address'):is_optional(),
+        ip = ts.boolean:describe('IP like address'):is_optional(),
+        braced = ts.boolean:describe('Have braces around address'):is_optional(),
+        quoted = ts.boolean:describe('Have quotes around address'):is_optional(),
+        empty = ts.boolean:describe('Empty address'):is_optional(),
+        backslash = ts.boolean:describe('Backslash in address'):is_optional(),
+        ['8bit'] = ts.boolean:describe('8 bit characters in address'):is_optional(),
+      },
+      user =  ts.string:describe('Parsed user part'):is_optional(),
+      name =  ts.string:describe('Displayed name'):is_optional(),
+      domain =  ts.string:describe('Parsed domain part'):is_optional(),
+    }
+  end
+  local function envelope_schema()
+    return ts.shape{
+      from_smtp = email_addr_schema():describe('SMTP from'):is_optional(),
+      recipients_smtp = ts.array_of(email_addr_schema()):describe('SMTP recipients'):is_optional(),
+      helo = ts.string:describe('SMTP Helo'):is_optional(),
+      hostname = ts.string:describe('Sender hostname'):is_optional(),
+      client_ip = ts.string:describe('Client ip'):is_optional(),
+      from_ip = ts.string:describe('Sender ip'):is_optional(),
+    }
+  end
+
+  return ts.shape{
+    headers = ts.array_of(headers_schema()),
+    parts = ts.array_of(part_schema()),
+    digest = ts.pattern(string.format('^%s$', string.rep('%x', 32)))
+        :describe('Message digest'),
+    newlines = ts.one_of({"cr", "lf", "crlf"}):describe('Newlines type'),
+    size = ts.integer:describe('Size of the message in bytes'),
+    envelope = envelope_schema()
+  }
+end
+
 return exports
