@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2018, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 Copyright (c) 2019, Carsten Rosenberg <c.rosenberg@heinlein-support.de>
 
 Licensed under the Apache License, Version 2.0 (the "License");
@@ -405,7 +405,7 @@ local function check_parts_match(task, rule)
           lua_util.debugm(rule.name, task, '%s: extension matched: |%s|%s|', rule.log_prefix, ext, ext2)
           return true
         elseif match_filter(task, rule, fname, rule.mime_parts_filter_regex, 'regex') then
-          lua_util.debugm(rule.name, task, '%s: filname regex matched', rule.log_prefix)
+          lua_util.debugm(rule.name, task, '%s: filename regex matched', rule.log_prefix)
           return true
         end
       end
@@ -459,14 +459,15 @@ local function check_parts_match(task, rule)
     end
 
     if rule.scan_all_mime_parts ~= false then
+      local is_part_checkable = (p:is_attachment() and (not p:is_image() or rule.scan_image_mime))
       if detected_ext then
         -- We know what to scan!
         local magic = lua_magic_types[detected_ext] or {}
 
-        if p:is_attachment() or magic.av_check ~= false then
+        if magic.av_check ~= false or is_part_checkable then
           return true
         end
-      elseif p:is_attachment() then
+      elseif is_part_checkable then
         -- Just rely on attachment property
         return true
       end
