@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2018, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -22,6 +22,7 @@ local rspamd_logger = require "rspamd_logger"
 local lua_meta = require "lua_meta"
 local rspamd_url = require "rspamd_url"
 local lua_util = require "lua_util"
+local lua_mime = require "lua_mime"
 local ucl = require "ucl"
 
 -- Define command line options
@@ -40,10 +41,12 @@ parser:mutex(
     parser:flag "-j --json"
           :description "JSON output",
     parser:flag "-U --ucl"
-          :description "UCL output"
+          :description "UCL output",
+    parser:flag "-M --messagepack"
+        :description "MessagePack output"
 )
 parser:flag "-C --compact"
-      :description "Use compactl format"
+      :description "Use compact format"
 parser:flag "--no-file"
       :description "Do not print filename"
 
@@ -76,6 +79,8 @@ extract:flag "-p --part"
        :description "Show part info"
 extract:flag "-s --structure"
        :description "Show structure info (e.g. HTML tags)"
+extract:flag "-i --invisible"
+       :description "Show invisible content for HTML parts"
 extract:option "-F --words-format"
        :description "Words format ('stem', 'norm', 'raw', 'full')"
        :argname("<type>")
@@ -115,7 +120,9 @@ urls:mutex(
     urls:flag "-t --tld"
         :description "Get TLDs only",
     urls:flag "-H --host"
-        :description "Get hosts only"
+        :description "Get hosts only",
+    urls:flag "-f --full"
+      :description "Show piecewise urls as processed by Rspamd"
 )
 
 urls:flag "-u --unique"
@@ -189,6 +196,22 @@ sign:option "-o --output"
     }
     :default 'message'
 
+local dump = parser:command "dump"
+                   :description "Dumps a raw message in different formats"
+dump:argument "file"
+    :description "File to process"
+    :argname "<file>"
+    :args "+"
+-- Duplicate format for convenience
+dump:mutex(
+    parser:flag "-j --json"
+          :description "JSON output",
+    parser:flag "-U --ucl"
+          :description "UCL output",
+    parser:flag "-M --messagepack"
+          :description "MessagePack output"
+)
+
 local function load_config(opts)
   local _r,err = rspamd_config:load_ucl(opts['config'])
 
@@ -234,16 +257,22 @@ local function maybe_print_fname(opts, fname)
   end
 end
 
+local function output_fmt(opts)
+  local fmt = 'json'
+  if opts.compact then fmt = 'json-compact' end
+  if opts.ucl then fmt = 'ucl' end
+  if opts.messagepack then fmt = 'msgpack' end
+
+  return fmt
+end
+
 -- Print elements in form
 -- filename -> table of elements
 local function print_elts(elts, opts, func)
   local fun = require "fun"
 
   if opts.json or opts.ucl then
-    local fmt = 'json'
-    if opts.compact then fmt = 'json-compact' end
-    if opts.ucl then fmt = 'ucl' end
-    io.write(ucl.to_format(elts, fmt))
+    io.write(ucl.to_format(elts, output_fmt(opts)))
   else
     fun.each(function(fname, elt)
 
@@ -263,6 +292,7 @@ end
 
 local function extract_handler(opts)
   local out_elts = {}
+  local tasks = {}
   local process_func
 
   if opts.words then
@@ -297,7 +327,6 @@ local function extract_handler(opts)
                 return string.format('%s:%s', k,v)
               end
             end, '', part:get_stats())))
-        table.insert(out, '\n')
       end
     end
   end
@@ -306,12 +335,15 @@ local function extract_handler(opts)
     if opts.part then
 
       if not opts.json and not opts.ucl then
+        local mtype,msubtype = part:get_type()
+        local det_mtype,det_msubtype = part:get_detected_type()
         table.insert(out,
-            rspamd_logger.slog('Mime Part: %s: %s/%s, filename: %s, size: %s',
+            rspamd_logger.slog('Mime Part: %s: %s/%s (%s/%s detected), filename: %s (%s detected ext), size: %s',
                 part:get_digest():sub(1,8),
-                ({part:get_type()})[1],
-                ({part:get_type()})[2],
+                mtype, msubtype,
+                det_mtype, det_msubtype,
                 part:get_filename(),
+                part:get_detected_ext(),
                 part:get_length()))
       end
     end
@@ -349,9 +381,9 @@ local function extract_handler(opts)
     end
 
     if opts.words then
-      local howw = opts['words_format'] or 'stem'
+      local how_words = opts['words_format'] or 'stem'
       table.insert(out_elts[fname], 'meta_words: ' ..
-          print_words(task:get_meta_words(howw), howw == 'full'))
+          print_words(task:get_meta_words(how_words), how_words == 'full'))
     end
 
     if opts.text or opts.html then
@@ -364,44 +396,73 @@ local function extract_handler(opts)
 
         if part and opts.text and not part:is_html() then
           maybe_print_text_part_info(part, out_elts[fname])
+          maybe_print_mime_part_info(mime_part, out_elts[fname])
+          if not opts.json and not opts.ucl then
+            table.insert(out_elts[fname], '\n')
+          end
+
           if opts.words then
-            local howw = opts['words_format'] or 'stem'
-            table.insert(out_elts[fname], print_words(part:get_words(howw),
-                howw == 'full'))
+            local how_words = opts['words_format'] or 'stem'
+            table.insert(out_elts[fname], print_words(part:get_words(how_words),
+                how_words == 'full'))
           else
             table.insert(out_elts[fname], tostring(part:get_content(how)))
           end
         elseif part and opts.html and part:is_html() then
           maybe_print_text_part_info(part, out_elts[fname])
+          maybe_print_mime_part_info(mime_part, out_elts[fname])
+          if not opts.json and not opts.ucl then
+            table.insert(out_elts[fname], '\n')
+          end
+
           if opts.words then
-            local howw = opts['words_format'] or 'stem'
-            table.insert(out_elts[fname], print_words(part:get_words(howw),
-                howw == 'full'))
+            local how_words = opts['words_format'] or 'stem'
+            table.insert(out_elts[fname], print_words(part:get_words(how_words),
+                how_words == 'full'))
           else
             if opts.structure then
               local hc = part:get_html()
               local res = {}
-              process_func = function(k, v)
-                return rspamd_logger.slog("%s = %s", k, v)
+              process_func = function(elt)
+                local fun = require "fun"
+                if type(elt) == 'table' then
+                  return table.concat(fun.totable(
+                      fun.map(
+                          function(t)
+                            return rspamd_logger.slog("%s", t)
+                          end,
+                          elt)), '\n')
+                else
+                  return rspamd_logger.slog("%s", elt)
+                end
               end
 
               hc:foreach_tag('any', function(tag)
-                local elt = {}
-                local ex = tag:get_extra()
-                elt.tag = tag:get_type()
-                if ex then
-                  elt.extra = ex
-                end
-                local content = tag:get_content()
-                if content then
-                  elt.content = content
-                end
-                table.insert(res, elt)
+                  local elt = {}
+                  local ex = tag:get_extra()
+                  elt.tag = tag:get_type()
+                  if ex then
+                    elt.extra = ex
+                  end
+                  local content = tag:get_content()
+                  if content then
+                    elt.content = tostring(content)
+                  end
+                  local style = tag:get_style()
+                  if style then
+                    elt.style = style
+                  end
+                  table.insert(res, elt)
               end)
-              out_elts[fname] = res
-            else
+              table.insert(out_elts[fname], res)
+            else -- opts.structure
               table.insert(out_elts[fname], tostring(part:get_content(how)))
             end
+            if opts.invisible then
+              local hc = part:get_html()
+              table.insert(out_elts[fname], string.format('invisible content: %s',
+                  tostring(hc:get_invisible())))
+            end
           end
         end
 
@@ -412,11 +473,12 @@ local function extract_handler(opts)
     end
 
     table.insert(out_elts[fname], "")
-
-    task:destroy() -- No automatic dtor
+    table.insert(tasks, task)
   end
 
   print_elts(out_elts, opts, process_func)
+  -- To avoid use after free we postpone tasks destruction
+  for _,task in ipairs(tasks) do task:destroy() end
 end
 
 local function stat_handler(opts)
@@ -512,6 +574,8 @@ local function urls_handler(opts)
         s = u:get_tld()
       elseif opts.host then
         s = u:get_host()
+      elseif opts.full then
+        s = rspamd_logger.slog('%s: %s', u:get_text(), u:to_table())
       else
         s = u:get_text()
       end
@@ -642,47 +706,6 @@ local function modify_handler(opts)
     return content
   end
 
-  local function do_append_footer(task, part, footer, is_multipart, out)
-    local newline_s = newline(task)
-    local tp = part:get_text()
-    local ct = 'text/plain'
-    local cte = 'quoted-printable'
-
-    if tp:is_html() then
-      ct = 'text/html'
-    end
-
-    if part:get_cte() == '7bit' then
-      cte = '7bit'
-    end
-
-    if is_multipart then
-      out[#out + 1] = string.format('Content-Type: %s; charset=utf-8%s'..
-          'Content-Transfer-Encoding: %s',
-          ct, newline_s, cte)
-      out[#out + 1] = ''
-    end
-
-    local content = tostring(tp:get_content('raw_utf') or '')
-    local double_nline = newline_s .. newline_s
-    local nlen = #double_nline
-    -- Hack, if part ends with 2 newline, then we append it after footer
-    if content:sub(-(nlen), nlen + 1) == double_nline then
-      content = string.format('%s%s',
-          content:sub(-(#newline_s), #newline_s + 1), -- content without last newline
-          footer)
-      out[#out + 1] = {rspamd_util.encode_qp(content,
-          80, task:get_newlines_type()), true}
-      out[#out + 1] = ''
-    else
-      content = content .. footer
-      out[#out + 1] = {rspamd_util.encode_qp(content,
-          80, task:get_newlines_type()), true}
-      out[#out + 1] = ''
-    end
-
-  end
-
   local text_footer, html_footer
 
   if opts['text_footer'] then
@@ -696,13 +719,12 @@ local function modify_handler(opts)
   for _,fname in ipairs(opts.file) do
     local task = load_task(opts, fname)
     local newline_s = newline(task)
-    local need_rewrite_ct = false
-    local parsed_ct
-    local seen_cte = false
-    local out = {}
+    local seen_cte
+
+    local rewrite = lua_mime.add_text_footer(task, html_footer, text_footer) or {}
+    local out = {} -- Start with headers
 
     local function process_headers_cb(name, hdr)
-
       for _,h in ipairs(opts['remove_header']) do
         if name:match(h) then
           return
@@ -723,16 +745,16 @@ local function modify_handler(opts)
         end
       end
 
-      if need_rewrite_ct then
+      if rewrite.need_rewrite_ct then
         if name:lower() == 'content-type' then
           local nct = string.format('%s: %s/%s; charset=utf-8',
-              'Content-Type', parsed_ct.type, parsed_ct.subtype)
+              'Content-Type', rewrite.new_ct.type, rewrite.new_ct.subtype)
           out[#out + 1] = nct
           return
         elseif name:lower() == 'content-transfer-encoding' then
-          seen_cte = true
           out[#out + 1] = string.format('%s: %s',
-              'Content-Transfer-Encoding', 'quoted-printable')
+              'Content-Transfer-Encoding', rewrite.new_cte or 'quoted-printable')
+          seen_cte = true
           return
         end
       end
@@ -740,50 +762,6 @@ local function modify_handler(opts)
       out[#out + 1] = hdr.raw:gsub('\r?\n?$', '')
     end
 
-    if html_footer or text_footer then
-      -- We need to take extra care about content-type and cte
-      local ct = task:get_header('Content-Type')
-      if ct then
-        ct = rspamd_util.parse_content_type(ct, task:get_mempool())
-      end
-
-      if ct then
-        if ct.type and ct.type == 'text' then
-          if ct.subtype then
-            if html_footer and (ct.subtype == 'html' or ct.subtype == 'htm') then
-              need_rewrite_ct = true
-            elseif text_footer and ct.subtype == 'plain' then
-              need_rewrite_ct = true
-            end
-          else
-            if text_footer then
-              need_rewrite_ct = true
-            end
-          end
-
-          parsed_ct = ct
-        end
-      else
-        local text_parts = task:get_text_parts()
-        if text_parts then
-
-          if #text_parts == 1 then
-            need_rewrite_ct = true
-            parsed_ct = {
-              type = 'text',
-              subtype = 'plain'
-            }
-          elseif #text_parts > 1 then
-            -- XXX: in fact, it cannot be
-            parsed_ct = {
-              type = 'multipart',
-              subtype = 'mixed'
-            }
-          end
-        end
-      end
-    end
-
     task:headers_foreach(process_headers_cb, {full = true})
 
     for _,h in ipairs(opts['add_header']) do
@@ -795,121 +773,40 @@ local function modify_handler(opts)
       end
     end
 
-    if not seen_cte and need_rewrite_ct then
+    if not seen_cte and rewrite.need_rewrite_ct then
       out[#out + 1] = string.format('%s: %s',
-          'Content-Transfer-Encoding', 'quoted-printable')
+          'Content-Transfer-Encoding', rewrite.new_cte or 'quoted-printable')
     end
 
     -- End of headers
-    --local eoh_pos = #out
     out[#out + 1] = ''
 
-    local boundaries = {}
-    local cur_boundary
-
-    for _,part in ipairs(task:get_parts()) do
-      local boundary = part:get_boundary()
-      if part:is_multipart() then
-        if cur_boundary then
-          out[#out + 1] = string.format('--%s',
-              boundaries[#boundaries])
-        end
-
-        boundaries[#boundaries + 1] = boundary or '--XXX'
-        cur_boundary = boundary
-
-        local rh = part:get_raw_headers()
-        if #rh > 0 then
-          out[#out + 1] = {rh, true}
-        end
-      elseif part:is_message() then
-        if boundary then
-          if cur_boundary and boundary ~= cur_boundary then
-            -- Need to close boundary
-            out[#out + 1] = string.format('--%s--%s',
-                boundaries[#boundaries], newline_s)
-            table.remove(boundaries)
-            cur_boundary = nil
-          end
-          out[#out + 1] = string.format('--%s',
-              boundary)
-        end
-
-        out[#out + 1] = {part:get_raw_headers(), true}
-      else
-        local append_footer = false
-        local skip_footer = part:is_attachment()
-
-        local parent = part:get_parent()
-        if parent then
-          local t,st = parent:get_type()
-
-          if t == 'multipart' and st == 'signed' then
-            -- Do not modify signed parts
-            skip_footer = true
-          end
-        end
-        if text_footer and part:is_text() then
-          local tp = part:get_text()
-
-          if not tp:is_html() then
-            append_footer = text_footer
-          end
-        end
-
-        if html_footer and part:is_text() then
-          local tp = part:get_text()
-
-          if tp:is_html() then
-            append_footer = html_footer
-          end
-        end
-
-        if boundary then
-          if cur_boundary and boundary ~= cur_boundary then
-            -- Need to close boundary
-            out[#out + 1] = string.format('--%s--%s',
-                boundaries[#boundaries], newline_s)
-            table.remove(boundaries)
-            cur_boundary = boundary
-          end
-          out[#out + 1] = string.format('--%s',
-              boundary)
-        end
-
-        io.flush()
-
-        if append_footer and not skip_footer then
-          do_append_footer(task, part, append_footer,
-              parent and parent:is_multipart(), out)
-        else
-          out[#out + 1] = {part:get_raw_headers(), true}
-          out[#out + 1] = {part:get_raw_content(), false}
-        end
+    if rewrite.out then
+      for _,o in ipairs(rewrite.out) do
+        out[#out + 1] = o
       end
-    end
-
-    -- Close remaining
-    local b = table.remove(boundaries)
-    while b do
-      out[#out + 1] = string.format('--%s--', b)
-      if #boundaries > 0 then
-        out[#out + 1] = ''
-      end
-      b = table.remove(boundaries)
+    else
+      out[#out + 1] = {task:get_rawbody(), false}
     end
 
     for _,o in ipairs(out) do
       if type(o) == 'string' then
         io.write(o)
         io.write(newline_s)
-      else
+      elseif type(o) == 'table' then
         io.flush()
-        o[1]:save_in_file(1)
+        if type(o[1]) == 'string' then
+          io.write(o[1])
+        else
+          o[1]:save_in_file(1)
+        end
 
         if o[2] then
           io.write(newline_s)
         end
+      else
+        o:save_in_file(1)
+        io.write(newline_s)
       end
     end
 
@@ -923,6 +820,11 @@ local function sign_handler(opts)
 
   local lua_dkim = require("lua_ffi").dkim
 
+  if not lua_dkim then
+    io.stderr:write('FFI support is required: please use LuaJIT or install lua-ffi')
+    os.exit(1)
+  end
+
   local sign_key
   if rspamd_util.file_exists(opts.key) then
     sign_key = lua_dkim.load_sign_key(opts.key, 'file')
@@ -971,6 +873,24 @@ local function sign_handler(opts)
   end
 end
 
+local function dump_handler(opts)
+  load_config(opts)
+  rspamd_url.init(rspamd_config:get_tld_path())
+
+  for _,fname in ipairs(opts.file) do
+    local task = load_task(opts, fname)
+
+    if opts.ucl or opts.json or opts.messagepack then
+      local ucl_object = lua_mime.message_to_ucl(task)
+      io.write(ucl.to_format(ucl_object, output_fmt(opts)))
+    else
+      task:get_content():save_in_file(1)
+    end
+
+    task:destroy() -- No automatic dtor
+  end
+end
+
 local function handler(args)
   local opts = parser:parse(args)
 
@@ -992,6 +912,8 @@ local function handler(args)
     modify_handler(opts)
   elseif command == 'sign' then
     sign_handler(opts)
+  elseif command == 'dump' then
+    dump_handler(opts)
   else
     parser:error('command %s is not implemented', command)
   end
